﻿// Copyright (c) 2021 Active Silicon
//
// A simple example showing how to interact with the
// ONVIF server running on the Harrier IP. The example creates
// a few overlays and exchanges VISCA messages with the camera.
// Make sure to change the IP address specified in the call to 
// FindService() to match that of your board.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.ServiceModel.Discovery;
using System.Collections.ObjectModel;

namespace HarrierIP_Ex1
{
   public class OnvifClient
   {
      public OnvifDevice.DeviceClient device;
      public OnvifMedia.MediaClient media;
      public OnvifDeviceIO.DeviceIOPortClient deviceio;

      static public string FindService(string ServiceIPv4)
      {
         var endPoint = new UdpDiscoveryEndpoint(DiscoveryVersion.WSDiscoveryApril2005);

         DiscoveryClient discoveryClient = new DiscoveryClient(endPoint);

         var findCriteria = new FindCriteria()
         {
            Duration = TimeSpan.FromSeconds(1),
            MaxResults = int.MaxValue
         };

         // Note, current spec https://www.onvif.org/specs/td/nvt/ONVIF-NVT-Definition-v210.pdf
         // states that we should look for a Type of tds:Device; the service should still contain 
         // NetworkVideoTransmitter in the <d:Types> declaration for backward compatibility.
         findCriteria.ContractTypeNames.Add(new System.Xml.XmlQualifiedName
             ("NetworkVideoTransmitter", @"http://www.onvif.org/ver10/network/wsdl"));

         var discoveryResponse = discoveryClient.Find(findCriteria);

         Collection<EndpointDiscoveryMetadata> onvifServices = discoveryResponse.Endpoints;

         foreach (var service in onvifServices)
         {
            foreach (var item in service.ListenUris)
            {
               if (item.OriginalString.Contains(ServiceIPv4))
                  return service.ListenUris[0].ToString();
            }
         }

         return null;
      }

      public void Init(string uri)
      {
         if (uri == null)
            throw new InvalidOperationException("Invalid URI.");

         System.ServiceModel.EndpointAddress remoteAddress = new System.ServiceModel.EndpointAddress(uri);

         var binding = new System.ServiceModel.Channels.CustomBinding("DeviceBinding");

         device = new OnvifDevice.DeviceClient(binding, remoteAddress);
         media = new OnvifMedia.MediaClient(binding, remoteAddress);
         deviceio = new OnvifDeviceIO.DeviceIOPortClient(binding, remoteAddress);
      }
   }

   public class Overlay
   {
      public const String CUSTOM = "Custom";
      public const String UPPERLEFT = "UpperLeft";
      public const String UPPERRIGHT = "UpperRight";
      public const String LOWERLEFT = "LowerLeft";
      public const String LOWERRIGHT = "LowerRight";

      // YUV White Text 
      private const float Y_TEXT = 1.0f;
      private const float U_TEXT = 0.5f;
      private const float V_TEXT = 0.5f;

      // YUV Black Background
      private const float Y_BACKGROUND = 0.0f;
      private const float U_BACKGROUND = 0.5f;
      private const float V_BACKGROUND = 0.5f;

      public OnvifMedia.OSDConfiguration osdConfig;

      public void SetPosition(String PosType, float PosX = 0, float PosY = 0)
      {
         osdConfig.Position = new OnvifMedia.OSDPosConfiguration();
         osdConfig.Position.Type = PosType;
         if (PosType == Overlay.CUSTOM)
         {
            osdConfig.Position.Pos = new OnvifMedia.Vector();
            osdConfig.Position.Pos.x = PosX;
            osdConfig.Position.Pos.xSpecified = true;
            osdConfig.Position.Pos.y = PosY;
            osdConfig.Position.Pos.ySpecified = true;
         }
      }
      public void InitTextOverlaySolidBackground(String text)
      {
         osdConfig.Type = OnvifMedia.OSDType.Text;
         osdConfig.TextString = new OnvifMedia.OSDTextConfiguration();
         osdConfig.TextString.PlainText = text;
         osdConfig.TextString.Type = "PlainText";
         osdConfig.TextString.FontColor = new OnvifMedia.OSDColor();
         osdConfig.TextString.FontColor.Color = new OnvifMedia.Color();
         osdConfig.TextString.FontColor.Color.X = Y_TEXT; 
         osdConfig.TextString.FontColor.Color.Y = U_TEXT;
         osdConfig.TextString.FontColor.Color.Z = V_TEXT;
         osdConfig.TextString.FontSize = 30;
         osdConfig.TextString.BackgroundColor = new OnvifMedia.OSDColor();
         osdConfig.TextString.BackgroundColor.Color = new OnvifMedia.Color();
         osdConfig.TextString.BackgroundColor.Color.X = Y_BACKGROUND; 
         osdConfig.TextString.BackgroundColor.Color.Y = U_BACKGROUND;
         osdConfig.TextString.BackgroundColor.Color.Z = V_BACKGROUND;
      }
      public void InitTextOverlayTransparent(String text = "")
      {
         osdConfig.Type = OnvifMedia.OSDType.Text;
         osdConfig.TextString = new OnvifMedia.OSDTextConfiguration();
         osdConfig.TextString.PlainText = text;
         osdConfig.TextString.Type = "PlainText";
         osdConfig.TextString.FontColor = new OnvifMedia.OSDColor();
         osdConfig.TextString.FontColor.Color = new OnvifMedia.Color();
         osdConfig.TextString.FontColor.Color.X = Y_BACKGROUND;
         osdConfig.TextString.FontColor.Color.Y = U_BACKGROUND;
         osdConfig.TextString.FontColor.Color.Z = V_BACKGROUND;
         osdConfig.TextString.FontSize = 30;
         osdConfig.TextString.BackgroundColor = new OnvifMedia.OSDColor();
         osdConfig.TextString.BackgroundColor.Color = new OnvifMedia.Color();
         osdConfig.TextString.BackgroundColor.TransparentSpecified = true;
         osdConfig.TextString.BackgroundColor.Transparent = 128;
      }
      public Overlay(string videoSourceToken)
      {
         osdConfig = new OnvifMedia.OSDConfiguration();
         osdConfig.VideoSourceConfigurationToken = new OnvifMedia.OSDReference();
         osdConfig.VideoSourceConfigurationToken.Value = videoSourceToken;
      }
   }

   public class CameraPort
   {
      public string portToken;
      public CameraPort(OnvifDeviceIO.DeviceIOPortClient deviceio)
      {
         var ports = deviceio.GetSerialPorts();
         if (ports.Length < 1)
            throw new InvalidOperationException("No serial ports found.");
         portToken = ports[0].token; // First seial port is that of the camera
      }
   }
   class Program
   {
      static void ClearOverlays(OnvifMedia.MediaClient media)
      {
         var videoSourceCfgs = media.GetVideoSourceConfigurations();

         // Delete existing overlays 
         foreach (var v in media.GetOSDs(videoSourceCfgs[0].token))
         {
            System.Xml.XmlElement[] unused = new System.Xml.XmlElement[0];
            media.DeleteOSD(v.token, ref unused);
         }
      }
      static void CreateOverlays(OnvifMedia.MediaClient media, string videoSourceToken)
      {
         var osd = new Overlay(videoSourceToken);
         osd.InitTextOverlaySolidBackground("Solid background.");
         osd.SetPosition(Overlay.UPPERLEFT);
         System.Xml.XmlElement[] unused = new System.Xml.XmlElement[0];
         media.CreateOSD(osd.osdConfig, ref unused);

         osd = new Overlay(videoSourceToken);
         osd.InitTextOverlayTransparent("Semi-transparent.");
         osd.SetPosition(Overlay.LOWERLEFT);
         media.CreateOSD(osd.osdConfig, ref unused);

         osd = new Overlay(videoSourceToken);
         osd.InitTextOverlayTransparent();
         osd.SetPosition(Overlay.LOWERRIGHT);
         osd.osdConfig.TextString.TimeFormat = "h:mm:ss tt";
         osd.osdConfig.TextString.Type = "Time";
         media.CreateOSD(osd.osdConfig, ref unused);

         osd = new Overlay(videoSourceToken);
         osd.SetPosition(Overlay.UPPERRIGHT);
         osd.osdConfig.Type = OnvifMedia.OSDType.Image;
         osd.osdConfig.Image = new OnvifMedia.OSDImgConfiguration();
         osd.osdConfig.Image.ImgPath = "AS_Logo.png";
         media.CreateOSD(osd.osdConfig, ref unused);

         osd = new Overlay(videoSourceToken);
         osd.SetPosition(Overlay.CUSTOM, 0.45f, 0.45f);
         osd.osdConfig.Type = OnvifMedia.OSDType.Image;
         osd.osdConfig.Image = new OnvifMedia.OSDImgConfiguration();
         osd.osdConfig.Image.ImgPath = "AS_Horizon.png";
         media.CreateOSD(osd.osdConfig, ref unused);
      }
      static bool ParseCmdLine(string [] args)
      {
         bool clear = true;

         System.Console.WriteLine("Run 'HarrierIP_Ex1 <IP address> 1' to run the example.");
         System.Console.WriteLine("Run 'HarrierIP_Ex1 <IP address> 0' to clear the overlays created by the example.");

         if (args.Length < 2)
         {
            throw new Exception("Please provide all required arguments.");
         }

         if (args.Length > 0)
         {
            if (Int32.Parse(args[1]) > 0)
               clear = false;
            else
               clear = true;
         }

         Console.WriteLine("IP address is: {0}", args[0]);

         return clear;
      }
      static void Main(string[] args)
      {
         try
         {
            var shouldCreateOverlays = ParseCmdLine(args);
            string ip = args[0];

            var onvifClient = new OnvifClient();

            string uri = OnvifClient.FindService(ip);

            if (uri == null)
            {
               System.Console.WriteLine("Unable to discover ONVIF service, assign URI (this is not an error).");
               uri = "http://" + ip + ":8000/onvif/device_service";
            }

            onvifClient.Init(uri);

            var videoSourceCfgs = onvifClient.media.GetVideoSourceConfigurations();
            if (videoSourceCfgs.Length < 1)
               throw new InvalidOperationException("No video source found.");

            // Remove all existing overlays 
            ClearOverlays(onvifClient.media);

            if (shouldCreateOverlays)
            {
               System.Console.WriteLine("The example will return after clearing the overlays.");
               return;
            }

            // Create new overlays for the demo
            CreateOverlays(onvifClient.media, videoSourceCfgs[0].token);

            // Swtich between Tele and Wide angle zoom as an example of exchanging
            // VISCA commands with the camera (Tamron camera).
            byte[] viscaCmdZoomTele = { 0x81, 0x01, 0x04, 0x07, 0x02, 0xFF };
            byte[] viscaCmdZoomWide = { 0x81, 0x01, 0x04, 0x07, 0x03, 0xFF };

            var cameraPort = new CameraPort(onvifClient.deviceio);
            var data = new OnvifDeviceIO.SerialData();

            System.Console.WriteLine("Setting camera to Tele zoom.");

            data.Item = viscaCmdZoomTele;
            onvifClient.deviceio.SendReceiveSerialCommand(cameraPort.portToken, ref data, "PT0M0.1S", "100", "");

            System.Threading.Thread.Sleep(2000);

            System.Console.WriteLine("Setting camera to Wide zoom.");

            data.Item = viscaCmdZoomWide;
            onvifClient.deviceio.SendReceiveSerialCommand(cameraPort.portToken, ref data, "PT0M0.1S", "100", "");
         } catch (SystemException e)
         {
            System.Console.WriteLine("Exception caught: " + e.Message);
         }
      }
   }
}
